import csv
import logging
from typing import Dict, Optional

from opennem.core.loader import load_data
from opennem.core.normalizers import clean_float, normalize_duid
from opennem.db import SessionLocal
from opennem.db.models.opennem import Facility
from opennem.importer.mms import mms_import

logger = logging.getLogger(__name__)


def import_mms_emissions() -> None:
    mms = mms_import()

    facility_poll_map = {
        # from wem
        "KWINANA_C5": 0.877,
        "KWINANA_C5": 0.877,
    }

    emission_map = []

    for station in mms:
        for facility in station.facilities:
            emission_factor = facility.emissions_factor_co2
            factor_source = "AEMO"

            if not emission_factor and facility.code in facility_poll_map.keys():
                emission_factor = facility_poll_map[facility.code]
                factor_source = "OpenNEM"

            emission_map.append(
                {
                    "station_name": station.name,
                    "network_id": "NEM",
                    "network_region": facility.network_region,
                    "facility_code": facility.code,
                    "emissions_factor_co2": facility.emissions_factor_co2,
                    "fueltech_id": "",
                    "emission_factor_source": factor_source,
                }
            )

    return emission_map


def import_emissions_csv() -> None:
    EMISSION_MAPS = [
        {"filename": "emission_factors.csv", "network": "WEM"},
        # {"filename": "nem_emissions.csv", "network": "NEM"},
    ]

    for m in EMISSION_MAPS:
        import_emissions_map(m["filename"])


def import_emissions_map(file_name: str) -> None:
    """Import emission factors from CSV files for each network
    the format of the csv file is

    station_name,network_id,network_region,facility_code,emissions_factor_co2,fueltech_id,emission_factor_source
    """
    session = SessionLocal()

    content = load_data(file_name, from_project=True, skip_loaders=True)

    csv_content = content.splitlines()
    csvreader = csv.DictReader(csv_content)

    for rec in csvreader:
        network_id = rec["network_id"]
        facility_code = normalize_duid(rec["facility_code"])
        emissions_intensity = clean_float(rec["emissions_factor_co2"])

        if not facility_code:
            logger.info("No emissions intensity for {}".format(facility_code))
            continue

        facility = (
            session.query(Facility)
            .filter_by(code=facility_code)
            .filter_by(network_id=network_id)
            .one_or_none()
        )

        if not facility:
            logger.info("No stored facility for {}".format(facility_code))
            continue

        facility.emissions_factor_co2 = emissions_intensity
        session.add(facility)
        logger.info("Updated {} to {}".format(facility_code, emissions_intensity))

    session.commit()

    return None


def check_emissions_map() -> None:
    content = load_data("opennem_emission_factors.csv", from_project=True, skip_loaders=True)
    mms_emissions = import_mms_emissions()

    def get_emissions_for_code(facility_code: str) -> Optional[Dict]:
        facility_lookup = list(
            filter(lambda x: x["facility_code"] == facility_code, mms_emissions)
        )

        if not facility_lookup or len(facility_lookup) < 1:
            logger.error("Could not find facility {} in MMS emmissions data".format(facility_code))
            return None

        facility = facility_lookup.pop()
        return facility

    csv_content = content.splitlines()
    csvreader = csv.DictReader(csv_content)

    csv_out = []

    for rec in csvreader:
        network_id = rec["network_id"]
        facility_code = normalize_duid(rec["facility_code"])
        emissions_intensity = clean_float(rec["emissions_factor_co2"])

        if network_id not in ["NEM"]:
            rec["emission_factor_source"] = "NPI"
            csv_out.append(rec)

            continue

        rec["emission_factor_source"] = ""

        mms_emission_record = get_emissions_for_code(facility_code)

        if not mms_emission_record:
            csv_out.append(rec)
            continue

        if emissions_intensity != mms_emission_record["emissions_factor_co2"]:
            logger.error(
                "Mismatch for {}: {} and {}".format(
                    facility_code, emissions_intensity, mms_emission_record["emissions_factor_co2"]
                )
            )

            if mms_emission_record["emissions_factor_co2"]:
                rec["emissions_factor_co2"] = mms_emission_record["emissions_factor_co2"]
            else:
                rec["emission_factor_source"] = "Lookup"

        rec["emission_factor_source"] = mms_emission_record["emission_factor_source"]
        csv_out.append(rec)

    fieldnames = [
        "network_id",
        "network_region",
        "facility_code",
        "station_name",
        "fueltech_id",
        "status_id",
        "emissions_factor_co2",
        "emission_factor_source",
    ]
    with open("emission_factors.csv", "w") as fh:
        csvwriter = csv.DictWriter(fh, fieldnames=fieldnames)
        csvwriter.writeheader()
        csvwriter.writerows(csv_out)


if __name__ == "__main__":
    check_emissions_map()



"""
OpenNEM Monitor emission factors for facilities

"""
import logging
from typing import List, Optional

from opennem.db import get_database_engine
from opennem.notifications.slack import slack_message
from opennem.schema.core import BaseConfig

logger = logging.getLogger("opennem.monitors.facility_seen")


class FacilityEmissionsRecord(BaseConfig):
    station_name: str
    network_id: str
    network_region: str
    facility_code: str
    emissions_factor_co2: Optional[float] = None
    fueltech_id: str


def get_facility_no_emission_factor(only_operating: bool = False) -> List[FacilityEmissionsRecord]:
    """Run this and it'll check if there are new facilities in
    that don't have emission factors
    """

    engine = get_database_engine()

    operating_filter = ""

    if only_operating:
        operating_filter = "and f.status_id='operating'"

    __query = """
        select
            s.name,
            f.network_id,
            f.network_region,
            f.code,
            f.emissions_factor_co2,
            f.fueltech_id
        from facility f
        left join station s on f.station_id = s.id
        left join fueltech ft on f.fueltech_id = ft.code
        where
            ft.renewable is False and
            ft.code not in ('imports', 'exports') and
            (f.emissions_factor_co2 is null or f.emissions_factor_co2 = 0)
            {operating_filter}
        order by network_region desc;
    """.format(
        operating_filter=operating_filter
    )

    with engine.connect() as c:
        logger.debug(__query)
        row = list(c.execute(__query))

    records: List[FacilityEmissionsRecord] = [
        FacilityEmissionsRecord(
            station_name=r[0],
            network_id=r[1],
            network_region=r[2],
            facility_code=r[3],
            emissions_factor_co2=r[4],
            fueltech_id=r[5],
        )
        for r in row
    ]

    return records


def alert_missing_emission_factors() -> None:
    missing_factors = get_facility_no_emission_factor()

    for rec in missing_factors:
        slack_message(
            "{} in {} {} with fueltech {} is missing factor".format(
                rec.station_name, rec.network_id, rec.network_region, rec.fueltech_id
            )
        )


if __name__ == "__main__":
    alert_missing_emission_factors()

    missing_factors = get_facility_no_emission_factor()

    for rec in missing_factors:
        print(
            "{} - {} in {} {} with fueltech {} is missing factor".format(
                rec.facility_code,
                rec.station_name,
                rec.network_id,
                rec.network_region,
                rec.fueltech_id,
            )
        )



# pylint: disable=no-member
"""
Remove old interconnect facilities and stations

Revision ID: 8746bdb27550
Revises: 6767d0102992
Create Date: 2020-12-24 14:56:00.999511

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "8746bdb27550"
down_revision = "6767d0102992"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "delete from station where id in (select distinct station_id from facility where fueltech_id in ('imports', 'exports'))"
    )


def downgrade() -> None:
    pass



# pylint: disable=no-member
"""
Update mv_wem_facility_power

Revision ID: 92c01037492d
Revises: e6bb59b27a23
Create Date: 2021-01-22 14:39:27.577816

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "92c01037492d"
down_revision = "e6bb59b27a23"
branch_labels = None
depends_on = None


stmt_drop = "drop materialized view if exists mv_wem_facility_power_30min cascade;"

stmt = """
CREATE MATERIALIZED VIEW mv_wem_facility_power_30min WITH (timescaledb.continuous) AS
select
    time_bucket('30 minutes', fs.trading_interval) as trading_interval,
    fs.facility_code,
    max(fs.generated) as facility_power
from facility_scada fs
where fs.network_id in ('WEM', 'APVI')
group by
    1,
    fs.facility_code;
"""


def upgrade() -> None:
    with op.get_context().autocommit_block():
        op.execute(stmt_drop)
        op.execute(stmt)


def downgrade() -> None:
    op.execute(stmt_drop)



from datetime import datetime
from typing import Dict, Generator

import scrapy

from opennem.pipelines.files import LinkExtract
from opennem.spiders.dirlisting import DirlistingSpider
from opennem.utils.dates import get_date_component


class WemCurrentSpider(scrapy.Spider):

    start_url = None

    url_params = {
        "day": get_date_component("%d"),
        "month": get_date_component("%m"),
        "year": get_date_component("%Y"),
    }

    def start_requests(self):

        cache_bust = datetime.now().strftime("%Y%M%d%H%M%S%f")

        if self.start_url:
            request_url = self.start_url.format(**self.url_params)
            request_url += "?{}".format(cache_bust)

            yield scrapy.Request(request_url)

    def parse(self, response) -> Generator[Dict, None, None]:
        yield {"content": response.text}


class WemHistoricSpider(DirlistingSpider):
    allowed_domains = ["wa.nemweb.com.au"]
    pipelines = set([LinkExtract])

    skip = 1

    # Archives tend to contain large zips of embedded zips so throttle
    # to limit memory use
    custom_settings = {
        "CONCURRENT_REQUESTS": 4,
        "CONCURRENT_ITEMS": 8,
    }



# pylint: disable=no-member
"""
create mv_interchange_power_nem_region

Revision ID: 4c35e2483eb7
Revises: d130be833a0c
Create Date: 2021-01-22 00:42:41.261473

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "4c35e2483eb7"
down_revision = "d130be833a0c"
branch_labels = None
depends_on = None

stmt_drop = "drop materialized view if exists mv_interchange_power_nem_region cascade;"

stmt = """
    CREATE MATERIALIZED VIEW mv_interchange_power_nem_region WITH (timescaledb.continuous) AS
    select
        time_bucket(INTERVAL '5 minutes', bs.trading_interval) as trading_interval,
        bs.network_id,
        bs.network_region,
        case when avg(bs.net_interchange) < 0 then
            avg(bs.net_interchange)
        else 0
        end as imports,
        case when avg(bs.net_interchange) > 0 then
            avg(bs.net_interchange)
        else 0
        end as exports
    from balancing_summary bs
    group by 1, 2, 3;
"""


def upgrade() -> None:
    with op.get_context().autocommit_block():
        op.execute(stmt_drop)
        op.execute(stmt)


def downgrade() -> None:
    op.execute(stmt_drop)



# pylint: disable=no-member
"""
Cascade station deletes

Revision ID: 6767d0102992
Revises: ae504bc24d95
Create Date: 2020-12-24 14:41:58.428764

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "6767d0102992"
down_revision = "ae504bc24d95"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
            alter table public.facility
                drop constraint if exists fk_station_status_code,
                drop constraint if exists fk_facility_station_code,
            add constraint fk_facility_station_code
                foreign key (station_id)
                references station(id)
                on delete cascade;
        """
    )


def downgrade() -> None:
    pass



import json
import logging
from typing import List

from opennem.core.facilitystations import facility_station_join_by_name
from opennem.core.loader import load_data
from opennem.db.models.opennem import Facility, Station
from opennem.exporter.encoders import OpenNEMJSONEncoder
from opennem.schema.opennem import StationSchema

from .aemo_gi import gi_import
from .aemo_rel import rel_import
from .mms import mms_import

logger = logging.getLogger("opennem.importer")

RECORD_MODEL_MAP = {
    "STATION": Station,
    "FACILITY": Facility,
}


def opennem_import_patches():
    """
    Reads the OpenNEM data source

    """

    opennem_records = load_data("opennem.csv", from_project=True)

    for rec in opennem_records:
        logger.debug(rec)

        if "record_type" not in rec:
            raise Exception("Invalid CSV: No record_type")

        record_type = rec["record_type"]

        if record_type not in RECORD_MODEL_MAP:
            raise Exception(
                "Invalid record type: {} is not a valid record type".format(
                    record_type
                )
            )

        record_model = RECORD_MODEL_MAP[record_type]

    return record_model


def record_diff(subject, comparitor):
    return {k: comparitor[k] for k in set(comparitor) - set(subject)}


def station_reindex(stations: List[dict]) -> List[dict]:
    """
    reindexes the opnnem formatted facilities from a list
    to a dict keyed by duid

    """
    stations_return = {}

    for station_code, station in stations.items():
        facs = []

        if "facilities" in station:
            facs = station.get("facilities")
            station.pop("facilities")

        fac_duids = [i["code"] for i in facs]

        if len(set(fac_duids)) != len(fac_duids):
            logger.info("Station {} has dupe duids".format(station_code))
            # raise Exception("error")

        station["facilities"] = {i.get("code"): i for i in facs}

        # stations_return[station_code] = StationSchema(**station)
        stations_return[station_code] = station

    return stations_return


def opennem_import():
    """
    This is the main method that overlays AEMO data and produces facilities

    """
    log = []

    nem_mms = station_reindex(load_data("mms.json", from_project=True))
    nem_rel = station_reindex(load_data("rel.json", from_project=True))
    nem_gi = station_reindex(load_data("nem_gi.json", True))
    registry = station_reindex(load_data("registry.json", True))

    opennem = nem_mms.copy()

    # REL
    for station_code, rel_station in nem_rel.items():
        if station_code not in opennem.keys():
            logger.info("REL: New record {}".format(station_code))
            opennem[station_code] = rel_station

        else:
            logger.info("REL: Existing record {}".format(station_code))
            om_station = opennem.get(station_code)

            for rel_facility_duid, rel_facility in rel_station[
                "facilities"
            ].items():

                if not rel_facility_duid:
                    continue

                if rel_facility_duid not in om_station["facilities"].keys():
                    logger.info(
                        " ==> Added duid {} to station ".format(
                            rel_facility_duid
                        )
                    )
                    om_station["facilities"][rel_facility_duid] = rel_facility
                    continue

                om_facility = om_station["facilities"][rel_facility_duid]

                if om_facility["status"]["code"] != "operating":
                    logger.info(
                        "REL: Set status for {} to {}".format(
                            rel_facility_duid,
                            "operating",
                        )
                    )
                    om_facility["status"]["code"] = "operating"
                    om_facility["status"]["label"] = "Operating"

                if rel_facility["fueltech"]:
                    logger.info(
                        "REL: Set fueltech for {} to {}".format(
                            rel_facility_duid,
                            rel_facility["fueltech"],
                        )
                    )
                    om_facility["fueltech"] = rel_facility["fueltech"]

                if (
                    rel_facility["capacity_registered"]
                    and rel_facility["capacity_registered"]
                    != om_facility["capacity_registered"]
                ):
                    logger.info(
                        "REL: Set capacity for %s to %s",
                        rel_facility_duid,
                        rel_facility["capacity_registered"],
                    )
                    om_facility["capacity_registered"] = rel_facility[
                        "capacity_registered"
                    ]

    # GI
    for station_code, gi_station in nem_gi.items():
        if (
            station_code not in opennem.keys()
            and not facility_station_join_by_name(gi_station["name"])
        ):
            logger.info("GI: New record {}".format(station_code))
            gi_station["created_by"] = "aemo_gi"
            opennem[station_code] = gi_station
            continue

        station_name = gi_station["name"]

        station_name_existing = list(
            filter(lambda x: x["name"] == gi_station["name"], opennem.values())
        )

        if len(station_name_existing) and facility_station_join_by_name(
            station_name
        ):
            station_code = station_name_existing[0]["code"]
            logger.info(
                "GI: found existing station we're joining to: {}".format(
                    station_code
                )
            )
        elif station_code not in opennem.keys():
            opennem[station_code] = gi_station

        logger.info("GI: Existing record {}".format(station_code))
        om_station = opennem.get(station_code)

        for gi_facility_duid, gi_facility in gi_station["facilities"].items():

            if not gi_facility_duid:
                continue

            if gi_facility_duid not in om_station["facilities"]:
                continue

            om_facility = om_station["facilities"][gi_facility_duid]

            if gi_facility_duid not in om_station["facilities"].keys():
                logger.info(
                    " ==> Added duid {} to station ".format(gi_facility_duid)
                )
                om_station["facilities"][gi_facility_duid] = gi_facility
                continue

            if (
                gi_facility["status"]
                and om_facility["status"]["code"] != gi_facility["status"]
            ):
                logger.info(
                    "GI Set status for {} to {}".format(
                        gi_facility_duid,
                        gi_facility["status"],
                    )
                )
                om_facility["status"] = gi_facility["status"]

            elif not om_facility["status"]:
                om_facility["status"] = gi_facility["status"]

            if not om_facility.get("fueltech", None):
                om_facility["fueltech"] = gi_facility["fueltech"]

    # registry
    for station_code, registry_station in registry.items():
        if station_code not in opennem.keys():
            if registry_station.get("state") == "WA":
                logger.info("Registry: New record {}".format(station_code))
                opennem[station_code] = registry_station

            continue

        om_station = opennem.get(station_code)

        lat = registry_station.get("lat", None)
        lng = registry_station.get("lng", None)

        om_station["lat"] = lat
        om_station["lng"] = lng

        if lat and lng:
            logger.info(
                "Registry: set lat and lng for {}".format(station_code)
            )

        for facility_code, registry_facility in registry_station.get(
            "facilities", {}
        ).items():

            if facility_code not in om_station["facilities"]:
                logger.info(
                    "Registry: {} has facility {} not in opennem".format(
                        station_code, facility_code
                    )
                )
                continue

            if (
                registry_facility["status"]
                and not om_station["facilities"][facility_code]["status"]
            ):
                logger.info(
                    "Registry: set status to {}".format(
                        registry_facility["status"]
                    )
                )
                om_station["facilities"][facility_code][
                    "status"
                ] = registry_station["status"]

            if registry_facility["fueltech"] and not om_station["facilities"][
                facility_code
            ].get("fueltech", None):
                om_station["facilities"][facility_code][
                    "fueltech"
                ] = registry_facility["fueltech"]

            logger.info(
                "registry: {} registry fueltech {} opennem fueltech {}".format(
                    facility_code,
                    registry_facility["fueltech"],
                    om_station["facilities"][facility_code].get(
                        "fueltech", None
                    ),
                )
            )

    for station_code, station_entry in opennem.items():
        facilities = [i for i in station_entry["facilities"].values()]
        opennem[station_code]["facilities"] = facilities

    with open("data/opennem.json", "w") as fh:
        json.dump(opennem, fh, indent=4, cls=OpenNEMJSONEncoder)

    stations = [StationSchema(**i) for i in list(opennem.values())]

    return stations


def opennem_export():
    pass


if __name__ == "__main__":
    opennem_import()



# pylint: disable=no-member
"""
create mv_region_emissions

Revision ID: 297112c824d0
Revises: 92c01037492d
Create Date: 2021-01-22 16:12:40.441310

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "297112c824d0"
down_revision = "92c01037492d"
branch_labels = None
depends_on = None

stmt_drop = "drop materialized view if exists mv_region_emissions;"

stmt = """
create materialized view mv_region_emissions as select
    f.trading_interval,
    f.network_region,
    case
        when sum(f.energy) > 0 then
            sum(f.emissions) / sum(f.energy)
        else 0
    end as emissions_per_kw
from mv_facility_all f
group by 1, f.network_region
order by 1 desc;
"""


def upgrade() -> None:
    with op.get_context().autocommit_block():
        op.execute(stmt_drop)
        op.execute(stmt)


def downgrade() -> None:
    op.execute(stmt_drop)



import logging
from typing import Any, Dict, List

from fastapi import APIRouter, HTTPException, Query
from starlette import status

from opennem.api.export.map import PriorityType, priority_from_name
from opennem.api.export.tasks import export_energy, export_power
from opennem.scheduler import huey
from opennem.utils.scrapyd import get_jobs, job_schedule_all

logger = logging.getLogger(__name__)

router = APIRouter()


@huey.task()
def export_energy_task(priority: PriorityType, latest: bool) -> None:
    export_energy(priority=priority, latest=latest)


@huey.task()
def export_power_task(priority: PriorityType, latest: bool) -> None:
    export_power(priority=priority, latest=latest)


@router.get("/scrapy/task")
def scrapy_task(
    name: str = Query(None, description="name of spider to queue")
) -> List[str]:
    return job_schedule_all(name)


@router.get("/scrapy/queue")
def scrapy_queue() -> Dict[str, Any]:
    resp = get_jobs()
    return resp


@router.get("/worker/task")
def worker_task(
    name: str = Query(None, description="type of task"),
    priority_name: str = Query("live", description="Task priority"),
    latest: bool = Query(False, description="Only run the latest"),
) -> Any:
    priority = priority_from_name(priority_name)

    task = None

    if name == "energy":
        task = export_energy_task.schedule(
            kwargs=dict(priority=priority, latest=latest), delay=1
        )
    elif name == "power":
        task = export_power_task.schedule(
            dict(priority=priority, latest=latest), delay=1
        )
    else:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Could not find task type",
        )

    return task.id


@router.get("/worker/queue")
def worker_queue() -> Any:
    res = huey.all_results().keys()

    return [i.decode("utf-8") for i in res]



# pylint: disable=no-member
"""
Fix APVI scada data for WEM

Revision ID: e79ef4fd3653
Revises: 72a817116434
Create Date: 2020-12-10 18:12:59.851256

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "e79ef4fd3653"
down_revision = "72a817116434"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "delete from facility_scada where facility_code = 'ROOFTOP_APVI_WA' and network_id='WEM'"
    )


def downgrade() -> None:
    pass



from typing import List

from scrapy import spiderloader
from scrapy.utils import project


def get_spiders() -> List[str]:
    """
    Get a list of spiders from scrapy

    """
    scrapy_settings = project.get_project_settings()
    spider_loader = spiderloader.SpiderLoader.from_settings(scrapy_settings)
    spiders = spider_loader.list()
    classes = [spider_loader.load(name) for name in spiders]

    return [str(i.name) for i in classes]



from opennem.pipelines.nem.opennem import NemwebUnitScadaOpenNEMStorePipeline
from opennem.spiders.nemweb import NemwebSpider


class NemwebLatestPriceSpider(NemwebSpider):
    # name = "au.nem.latest.price"
    start_url = "http://www.nemweb.com.au/Reports/CURRENT/Dispatchprices_PRE_AP/"
    limit = 2

    pipelines_extra = set([NemwebUnitScadaOpenNEMStorePipeline])


class NemwebCurrentPriceSpider(NemwebSpider):
    # name = "au.nem.current.price"
    start_url = "http://www.nemweb.com.au/Reports/CURRENT/Dispatchprices_PRE_AP/"
    limit = 0

    pipelines_extra = set([NemwebUnitScadaOpenNEMStorePipeline])


class NemwebArchivePriceSpider(NemwebSpider):
    # name = "au.nem.archive.price"
    start_url = "http://www.nemweb.com.au/Reports/ARCHIVE/Dispatchprices_PRE_AP/"
    limit = 0

    pipelines_extra = set([NemwebUnitScadaOpenNEMStorePipeline])

    # Archives tend to contain large zips of embedded zips so throttle
    # to limit memory use
    custom_settings = {
        "CONCURRENT_REQUESTS": 1,
        "CONCURRENT_ITEMS": 1,
    }



"""
    Module to map parameters between versions of opennem
"""


def map_compat_fueltech(fueltech: str) -> str:
    """
        Map old opennem fueltechs to new fueltechs

    """
    if fueltech == "brown_coal":
        return "coal_brown"

    if fueltech == "black_coal":
        return "coal_black"

    if fueltech == "solar":
        return "solar_utility"

    if fueltech == "biomass":
        return "bioenergy_biomass"

    return fueltech


def map_compat_network_region(network_region: str) -> str:
    """
        Map network regions from old to new

        Note that network regions aren't really geos
        and there are networks within geos like DKIS (NT)
        and NWIS (WA) that need to retain their network_region
    """
    if not network_region or not type(network_region) is str:
        return network_region

    network_region = network_region.strip()

    if network_region == "WA1":
        return "WEM"

    return network_region


def map_compat_facility_state(facility_state: str) -> str:
    """
        Maps a current v1 version opennem facility_registry.json state
        to a v3 state

    """
    state = facility_state.lower().strip()

    if state == "commissioned":
        return "operating"

    if state == "decommissioned":
        return "retired"

    return state



# pylint: disable=no-member
"""
Update mv_facility_energy_hour

Revision ID: 3f6867b6de7e
Revises: 9c7482a0fc9a
Create Date: 2021-01-19 17:11:21.222167

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "3f6867b6de7e"
down_revision = "9c7482a0fc9a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
    drop view if exists mv_facility_energy_hour cascade;

    CREATE OR REPLACE VIEW mv_facility_energy_hour WITH (timescaledb.continuous, timescaledb.refresh_interval = '1h', timescaledb.refresh_lag='1h') AS
    select
        time_bucket('1 hour', fs.trading_interval) as trading_interval,
        fs.facility_code,
        fs.network_id,
        energy_sum(fs.generated, '1 hour') * interval_size('1 hour', count(fs.generated)) as energy
    from facility_scada fs
    group by
        1, 2, 3;
    """
    )


def downgrade() -> None:
    pass



from __future__ import annotations

from datetime import date, datetime, timedelta
from decimal import Decimal
from typing import Any, List, Optional, Tuple, Union

from pydantic import validator

from opennem.api.time import human_to_interval
from opennem.core.compat.utils import translate_id_v3_to_v2
from opennem.core.fueltechs import map_v3_fueltech
from opennem.schema.core import BaseConfig
from opennem.schema.network import NetworkSchema
from opennem.schema.time import TimeIntervalAPI, TimePeriodAPI
from opennem.settings import settings
from opennem.utils.interval import get_human_interval
from opennem.utils.numbers import sigfig_compact
from opennem.utils.timezone import get_current_timezone


def chop_microseconds(dt: datetime) -> datetime:
    if not dt.microsecond:
        return dt

    return dt - timedelta(microseconds=dt.microsecond)


def optionaly_lowercase_string(value: str) -> str:
    """Read from settings if we want output schema string
    values to be lowercased or not and perform"""
    if settings.schema_output_lowercase_strings:
        value = value.lower()

    return value


def number_output(n: Union[float, int, None]) -> Optional[Union[float, int, None]]:
    """Format numbers for data series outputs"""
    if n is None:
        return None

    if n == 0:
        return 0

    return sigfig_compact(n)


def data_validate(values: List[Union[float, int, None]]) -> List[Union[float, int, None]]:
    """Validate and format list of numeric data values"""
    return list(
        map(
            number_output,
            values,
        )
    )


def optionally_parse_string_datetime(
    value: Optional[Union[str, datetime, date]] = None
) -> Optional[Union[str, datetime, date]]:
    if not value:
        return value

    if isinstance(value, str):
        return datetime.fromisoformat(value)

    return value


def get_data_id(
    network: NetworkSchema,
) -> str:
    """Method to build the data id for a stat export"""
    id_components = []

    if settings.schema_output_id_country:
        id_components.append(network.country)

    id_str = ".".join([i.lower() for i in id_components if i])

    return id_str


class OpennemDataHistory(BaseConfig):
    start: datetime
    last: datetime
    interval: str
    data: List

    # validators
    _data_valid = validator("data", allow_reuse=True, pre=True)(data_validate)

    def values(self) -> List[Tuple[datetime, float]]:
        interval_obj = get_human_interval(self.interval)
        interval_def = human_to_interval(self.interval)
        inclusive = False
        dt = self.start

        if interval_def.interval < 1440:
            inclusive = True

        # return as list rather than generate
        timeseries_data = []

        # rewind back one interval
        if inclusive:
            dt -= interval_obj

        for v in self.data:
            dt = dt + interval_obj
            timeseries_data.append((dt, v))

        # @TODO do some sanity checking here
        # if dt != self.last:
        #     raise Exception(
        #         "Mismatch between start, last and interval size. Got {} and {}".format(
        #             dt, self.last
        #         )
        #     )

        return timeseries_data


class OpennemData(BaseConfig):
    id: Optional[str]
    type: Optional[str]
    fuel_tech: Optional[str]

    network: Optional[str]
    region: Optional[str]
    data_type: str
    code: str = ""
    units: str

    interval: Optional[TimeIntervalAPI]
    period: Optional[TimePeriodAPI]

    history: OpennemDataHistory
    forecast: Optional[OpennemDataHistory]

    # validators
    _id_lowercase = validator("id", allow_reuse=True, pre=True)(optionaly_lowercase_string)
    _network_lowercase = validator("network", allow_reuse=True, pre=True)(
        optionaly_lowercase_string
    )
    _region_lowercase = validator("region", allow_reuse=True, pre=True)(optionaly_lowercase_string)
    _code_lowercase = validator("code", allow_reuse=True, pre=True)(optionaly_lowercase_string)

    # conveniance methods
    def id_v2(self) -> str:
        return translate_id_v3_to_v2(self.id)

    def fueltech_v2(self) -> Optional[str]:
        if self.fuel_tech:
            return map_v3_fueltech(self.fuel_tech)

        return None


class OpennemDataSet(BaseConfig):
    type: Optional[str]
    version: Optional[str]
    network: Optional[str]
    code: str
    region: Optional[str]
    created_at: Optional[datetime]

    data: List[OpennemData]

    def append_set(self, subject_set: Optional[OpennemDataSet] = None):
        if not subject_set:
            return None

        if not subject_set.data:
            return None

        if not isinstance(subject_set.data, list):
            return None

        if len(subject_set.data):
            self.data += subject_set.data

        return None

    def get_id(self, id: str) -> Optional[OpennemData]:
        _ds = list(filter(lambda x: x.id == id, self.data))

        if len(_ds) < 1:
            return None

        return _ds.pop()

    # validators
    _version_fromstr = validator("created_at", allow_reuse=True, pre=True)(
        optionally_parse_string_datetime
    )

    _created_at_trim = validator("created_at", allow_reuse=True, pre=True)(chop_microseconds)
    _network_lowercase = validator("network", allow_reuse=True, pre=True)(
        optionaly_lowercase_string
    )
    _region_lowercase = validator("region", allow_reuse=True, pre=True)(optionaly_lowercase_string)
    _code_lowercase = validator("code", allow_reuse=True, pre=True)(optionaly_lowercase_string)


class RegionFlowResult(BaseConfig):
    interval: datetime
    flow_from: str
    flow_to: str
    generated: Optional[float]
    flow_from_energy: Optional[float]
    flow_to_energy: Optional[float]


class RegionFlowEmissionsResult(BaseConfig):
    interval: datetime
    flow_from: str
    flow_to: str
    energy: float
    flow_from_emissions: Optional[float]
    flow_to_emissions: Optional[float]
    flow_from_intensity: Optional[float]
    flow_to_intensity: Optional[float]


class DataQueryResult(BaseConfig):
    interval: datetime
    result: Union[float, int, None, Decimal]
    group_by: Optional[str]


class ScadaDateRange(BaseConfig):
    start: datetime
    end: datetime
    network: Optional[NetworkSchema]

    def _get_value_localized(self, field_name: str = "start") -> Any:
        timezone = get_current_timezone()
        date_aware = getattr(self, field_name)

        if self.network:
            timezone = self.network.get_timezone()

        date_aware = date_aware.astimezone(timezone)

        return date_aware

    def get_start_year(self) -> int:
        return self.start.year

    def get_start(self) -> datetime:
        return self._get_value_localized("start")

    def get_end(self) -> datetime:
        return self._get_value_localized("end")

    def get_start_sql(self, as_date: bool = False) -> str:
        return "'{}'".format(self.get_start() if not as_date else self.get_start().date())

    def get_end_sql(self, as_date: bool = False) -> str:
        return "'{}'".format(self.get_end() if not as_date else self.get_end().date())



# pylint: disable=no-member
"""
idx_mv_facility_unique unique index on mv_facility_all

Revision ID: 9ce2abfe97e8
Revises: e161acce7727
Create Date: 2021-01-18 18:39:19.623430

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "9ce2abfe97e8"
down_revision = "e161acce7727"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "CREATE UNIQUE INDEX IF NOT EXISTS idx_mv_facility_unique ON mv_facility_all (trading_interval, network_id, code);"
    )


def downgrade() -> None:
    op.execute("drop index idx_mv_facility_unique;")



# pylint: disable=no-member
"""
create idx_mv_region_emissions_unique

Revision ID: 6100cbbcaf7e
Revises: 297112c824d0
Create Date: 2021-01-22 16:15:38.204480

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "6100cbbcaf7e"
down_revision = "297112c824d0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "CREATE UNIQUE INDEX IF NOT EXISTS idx_mv_region_emissions_unique ON mv_region_emissions (trading_interval, network_region);"
    )


def downgrade() -> None:
    op.execute("drop index if exists idx_mv_region_emissions_unique")



# pylint: disable=no-member
"""
Facility scada remove redundant interconnect data

Revision ID: 4569d3c0509e
Revises: 000e7fdab7e2
Create Date: 2020-12-24 15:09:21.715722

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "4569d3c0509e"
down_revision = "000e7fdab7e2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "delete from facility_scada where facility_code in ('V-SA-3', 'V-SA-4', 'N-Q-MNSP1-3', 'N-Q-MNSP1-4', 'NSW1-QLD1-3', 'NSW1-QLD1-4', 'V-S-MNSP1-3', 'V-S-MNSP1-4', 'T-V-MNSP1-3', 'T-V-MNSP1-4', 'VIC1-NSW1-3', 'VIC1-NSW1-4')"
    )


def downgrade() -> None:
    pass



from textwrap import dedent
from typing import List, Optional

from sqlalchemy import sql
from sqlalchemy.sql.elements import TextClause

from opennem.api.stats.controllers import networks_to_in
from opennem.schema.dates import TimeSeries
from opennem.schema.network import NetworkNEM, NetworkSchema, NetworkWEM
from opennem.schema.stats import StatTypes


def weather_observation_query(time_series: TimeSeries, station_codes: List[str]) -> str:

    if time_series.interval.interval > 1440:
        # @TODO replace with mv
        __query = """
        select
            date_trunc('{trunc}', t.observation_time at time zone '{tz}') as observation_month,
            t.station_id,
            avg(t.temp_avg),
            min(t.temp_min),
            max(t.temp_max)
        from
            (
                select
                    time_bucket_gapfill('1 day', observation_time) as observation_time,
                    fs.station_id,
                    avg(fs.temp_air) as temp_avg,
                    min(fs.temp_air) as temp_min,
                    max(fs.temp_air) as temp_max
                from bom_observation fs
                where
                    fs.station_id in ({station_codes}) and
                    fs.observation_time <= '{date_end}' and
                    fs.observation_time >= '{date_start}'
                group by 1, 2
            ) as t
        group by 1, 2;
        """.format(
            trunc=time_series.interval.trunc,
            tz=time_series.network.timezone_database,
            station_codes=",".join(["'{}'".format(i) for i in station_codes]),
            date_start=time_series.get_range().start,
            date_end=time_series.get_range().end,
        )

    else:
        __query = """
        select
            time_bucket_gapfill('{interval_sql}', observation_time) as ot,
            fs.station_id as station_id,
            avg(fs.temp_air) as temp_air,
            -- case  @TODO case on temp_max temp_min
            -- end case
            min(fs.temp_air) as temp_min,
            max(fs.temp_air) as temp_max
        from bom_observation fs
        where
            fs.station_id in ({station_codes}) and
            fs.observation_time <= '{date_end}' and
            fs.observation_time >= '{date_start}'
        group by 1, 2;
        """.format(
            interval_sql=time_series.interval.interval_sql,
            station_codes=",".join(["'{}'".format(i) for i in station_codes]),
            date_start=time_series.get_range().start,
            date_end=time_series.get_range().end,
        )

    return dedent(__query)


def interconnector_power_flow(time_series: TimeSeries, network_region: str) -> str:
    """Get interconnector region flows using materialized view"""
    ___query = """
        select
            trading_interval at time zone '{timezone}' as trading_interval,
            fs.network_region,
            fs.imports,
            fs.exports
        from mv_interchange_power_nem_region fs
        where
            fs.network_id = '{network_id}' and
            fs.network_region= '{region}' and
            fs.trading_interval <= '{date_end}' and
            fs.trading_interval >= '{date_start}'
        order by trading_interval asc;
    """.format(
        timezone=time_series.network.timezone_database,
        network_id=time_series.network.code,
        region=network_region,
        date_start=time_series.get_range().start,
        date_end=time_series.get_range().end,
    )

    return dedent(___query)


def country_stats_query(stat_type: StatTypes, country: str = "au") -> TextClause:
    __query = sql.text(
        dedent(
            """
                select
                    s.stat_date,
                    s.value,
                    s.stat_type
                from stats s
                where s.stat_type = :stat_type and s.country= :country
                order by s.stat_date desc
           """
        )
    ).bindparams(
        stat_type=str(stat_type),
        country=country,
    )

    return __query


def price_network_query(
    time_series: TimeSeries,
    group_field: str = "bs.network_id",
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:

    if not networks_query:
        networks_query = [time_series.network]

    if time_series.network not in networks_query:
        networks_query.append(time_series.network)

    __query = """
        select
            time_bucket_gapfill('{trunc}', bs.trading_interval) as trading_interval,
            {group_field},
            avg(bs.price) as price
        from balancing_summary bs
        where
            bs.trading_interval <= '{date_max}' and
            bs.trading_interval > '{date_min}' and
            {network_query}
            {network_region_query}
            1=1
        group by 1, 2
        order by 1 desc
    """

    timezone = time_series.network.timezone_database
    network_region_query = ""

    if network_region:
        network_region_query = f"bs.network_region='{network_region}' and "
        group_field = "bs.network_region"

    network_query = "bs.network_id IN ({}) and ".format(networks_to_in(networks_query))

    if len(networks_query) > 1:
        group_field = "'AU'"

    date_max = time_series.get_range().end
    date_min = time_series.get_range().start

    query = dedent(
        __query.format(
            network_query=network_query,
            trunc=time_series.interval.interval_sql,
            network_region_query=network_region_query,
            timezone=timezone,
            date_max=date_max,
            date_min=date_min,
            group_field=group_field,
        )
    )

    return query


def network_demand_query(
    time_series: TimeSeries,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:
    if not networks_query:
        networks_query = [time_series.network]

    if time_series.network not in networks_query:
        networks_query.append(time_series.network)

    __query = """
    select
        trading_interval at time zone '{timezone}',
        network_id,
        max(demand_total) as demand
    from balancing_summary bs
    where
        bs.trading_interval <= '{date_max}' and
        bs.trading_interval >= '{date_min}' and
        {network_query}
        {network_region_query}
        1=1
    group by
        1, {groups_additional}
    order by 1 asc;
    """

    group_keys = ["network_id"]
    network_region_query = ""

    if network_region:
        group_keys.append("network_region")
        network_region_query = f"bs.network_region = '{network_region}' and "

    groups_additional = ", ".join(group_keys)

    network_query = "bs.network_id IN ({}) and ".format(networks_to_in(networks_query))

    date_max = time_series.get_range().end
    date_min = time_series.get_range().start

    query = __query.format(
        timezone=time_series.network.timezone_database,
        date_max=date_max,
        date_min=date_min,
        network_id=time_series.network.code,
        network_query=network_query,
        network_region_query=network_region_query,
        groups_additional=groups_additional,
    )

    return dedent(query)


def power_network_fueltech_query(
    time_series: TimeSeries,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:
    """Query power stats"""

    if not networks_query:
        networks_query = [time_series.network]

    if time_series.network not in networks_query:
        networks_query.append(time_series.network)

    __query = """
    select
        time_bucket_gapfill('{trunc}', fs.trading_interval) AS trading_interval,
        ft.code as fueltech_code,
        sum(fs.generated) as facility_power
    from facility_scada fs
    join facility f on fs.facility_code = f.code
    join fueltech ft on f.fueltech_id = ft.code
    where
        fs.is_forecast is False and
        f.fueltech_id is not null and
        f.fueltech_id not in ({fueltechs_exclude}) and
        {network_query}
        {network_region_query}
        fs.trading_interval <= '{date_max}' and
        fs.trading_interval > '{date_min}'
        {fueltech_filter}
    group by 1, 2
    """

    network_region_query: str = ""
    fueltech_filter: str = ""
    wem_apvi_case: str = ""
    timezone: str = time_series.network.timezone_database

    fueltechs_excluded = ["exports", "imports"]

    if NetworkNEM in networks_query:
        fueltechs_excluded.append("solar_rooftop")

    if network_region:
        network_region_query = f"f.network_region='{network_region}' and "

    if NetworkWEM in networks_query:
        # silly single case we'll refactor out
        # APVI network is used to provide rooftop for WEM so we require it
        # in country-wide totals
        wem_apvi_case = "or (f.network_id='APVI' and f.network_region='WEM')"

    network_query = "(f.network_id IN ({}) {}) and ".format(
        networks_to_in(networks_query), wem_apvi_case
    )

    date_max = time_series.get_range().end
    date_min = time_series.get_range().start

    fueltechs_exclude = ", ".join("'{}'".format(i) for i in fueltechs_excluded)

    query = dedent(
        __query.format(
            network_query=network_query,
            trunc=time_series.interval.interval_sql,
            network_region_query=network_region_query,
            timezone=timezone,
            date_max=date_max,
            date_min=date_min,
            fueltech_filter=fueltech_filter,
            wem_apvi_case=wem_apvi_case,
            fueltechs_exclude=fueltechs_exclude,
        )
    )

    return query


def power_network_rooftop_query(
    time_series: TimeSeries,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
    forecast: bool = False,
) -> str:
    """Query power stats"""

    if not networks_query:
        networks_query = [time_series.network]

    if time_series.network not in networks_query:
        networks_query.append(time_series.network)

    __query = """
        select
            time_bucket_gapfill('30 minutes', fs.trading_interval)  AS trading_interval,
            ft.code as fueltech_code,
            coalesce(sum(fs.generated), 0) as facility_power
        from facility_scada fs
        join facility f on fs.facility_code = f.code
        join fueltech ft on f.fueltech_id = ft.code
        where
            fs.is_forecast is {forecast} and
            f.fueltech_id = 'solar_rooftop' and
            {network_query}
            {network_region_query}
            fs.trading_interval <= '{date_max}' and
            fs.trading_interval > '{date_min}'
        group by 1, 2
        order by 1 asc
    """

    network_region_query: str = ""
    wem_apvi_case: str = ""
    timezone: str = time_series.network.timezone_database

    if network_region:
        network_region_query = f"f.network_region='{network_region}' and "

    if NetworkWEM in networks_query:
        # silly single case we'll refactor out
        # APVI network is used to provide rooftop for WEM so we require it
        # in country-wide totals
        wem_apvi_case = "or (f.network_id='APVI' and f.network_region='WEM')"

    network_query = "(f.network_id IN ({}) {}) and ".format(
        networks_to_in(networks_query), wem_apvi_case
    )

    date_max = time_series.get_range().end
    date_min = time_series.get_range().start

    query = dedent(
        __query.format(
            network_query=network_query,
            network_region_query=network_region_query,
            timezone=timezone,
            date_max=date_max,
            date_min=date_min,
            forecast=str(forecast),
        )
    )

    return query


"""
Energy Queries
"""


def energy_network_fueltech_query(
    time_series: TimeSeries,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:
    """
    Get Energy for a network or network + region
    based on a year
    """

    if not networks_query:
        networks_query = [time_series.network]

    if time_series.network not in networks_query:
        networks_query.append(time_series.network)

    if time_series.interval.interval > 1440:
        __query = """
        select
            date_trunc('{trunc}', t.trading_day) as trading_month,
            t.fueltech_id,
            coalesce(sum(t.fueltech_energy) / 1000 , 0) as fueltech_energy,
            coalesce(sum(t.fueltech_market_value), 0) as fueltech_market_value,
            coalesce(sum(t.fueltech_emissions), 0) as fueltech_emissions
        from
            (select
                time_bucket_gapfill('1 day', t.ti_day_aest) as trading_day,
                t.fueltech_id,
                sum(t.energy) as fueltech_energy,
                sum(t.market_value) as fueltech_market_value,
                sum(t.emissions) as fueltech_emissions
            from mv_facility_all t
            where
                t.ti_day_aest <= '{date_max}' and
                t.ti_day_aest >= '{date_min}' and
                t.fueltech_id not in ('imports', 'exports') and
                {network_query}
                {network_region_query}
                1=1
            group by 1, 2) as t
        group by 1, 2
        order by
            1 desc;
        """
    else:
        __query = """
        select
            t.ti_{trunc_name} as trading_day,
            t.fueltech_id,
            coalesce(sum(t.energy) / 1000, 0) as fueltech_energy,
            coalesce(sum(t.market_value), 0) as fueltech_market_value,
            coalesce(sum(t.emissions), 0) as fueltech_emissions
        from mv_facility_all t
        where
            t.trading_interval <= '{date_max}' and
            t.trading_interval >= '{date_min}' and
            t.fueltech_id not in ('imports', 'exports') and
            {network_query}
            {network_region_query}
            1=1
        group by 1, 2
        order by
            trading_day desc;
        """

    network_region_query = ""
    date_range = time_series.get_range()

    if network_region:
        network_region_query = f"t.network_region='{network_region}' and"

    networks_list = networks_to_in(networks_query)
    network_query = "t.network_id IN ({}) and ".format(networks_list)

    trunc_name = "{}_{}".format(
        time_series.interval.trunc, time_series.network.timezone_database
    ).lower()

    query = dedent(
        __query.format(
            trunc=date_range.interval.trunc,
            trunc_name=trunc_name,
            date_min=date_range.start,
            date_max=date_range.end,
            network_query=network_query,
            network_region_query=network_region_query,
        )
    )

    return query


def energy_network_region_net_flows_query(
    time_series: TimeSeries,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:
    """
    Get Energy for a network or network + region
    based on a year
    """

    if not networks_query:
        networks_query = [time_series.network]

    if time_series.network not in networks_query:
        networks_query.append(time_series.network)

    __query = """
    select
        date_trunc('{trunc}', t.trading_interval at time zone '{timezone}') as trading_day,
        sum(t.imports_avg) / 1000,
        sum(t.exports_avg) / 1000
    from mv_interchange_energy_nem_region t
    where
        t.trading_interval <= '{date_max}' and
        t.trading_interval >= '{date_min}' and
        {network_query}
        {network_region_query}
        1=1
    group by 1
    order by 1 desc

    """
    timezone = time_series.network.timezone_database
    network_region_query = ""
    date_range = time_series.get_range()

    if network_region:
        network_region_query = f"t.network_region='{network_region}' and"

    networks_list = networks_to_in(networks_query)
    network_query = "t.network_id IN ({}) and ".format(networks_list)

    query = dedent(
        __query.format(
            timezone=timezone,
            trunc=date_range.interval.trunc,
            date_min=date_range.start,
            date_max=date_range.end,
            network_query=network_query,
            network_region_query=network_region_query,
        )
    )

    return query


def energy_network_interconnector_emissions_query(
    time_series: TimeSeries,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:
    """
    Get emissions for a network or network + region
    based on a year
    """

    if not networks_query:
        networks_query = [time_series.network]

    if time_series.network not in networks_query:
        networks_query.append(time_series.network)

    __query = """
    select
        t.trading_interval at time zone '{timezone}' as trading_interval,
        t.flow_from,
        t.flow_to,
        t.flow_energy as energy,
        t.flow_from_emissions,
        t.flow_to_emissions
    from vw_region_flow_emissions t
    where
        t.trading_interval <= '{date_max}' and
        t.trading_interval >= '{date_min}' and
        {network_region_query}
        1=1
    order by 1 desc

    """

    timezone = time_series.network.timezone_database
    network_region_query = ""
    date_range = time_series.get_range()

    if network_region:
        network_region_query = f"""
            (t.flow_from='{network_region}' or t.flow_to='{network_region}') and
        """

    query = dedent(
        __query.format(
            timezone=timezone,
            # trunc=date_range.interval.trunc,
            date_min=date_range.start,
            date_max=date_range.end,
            network_region_query=network_region_query,
        )
    )

    return query



from typing import Optional

from opennem.core.loader import load_data

STATION_NAME_CODE_MAP = load_data("station_name_code_map.json")


def station_name_code_map(station_name: str) -> Optional[str]:
    if station_name.strip() in STATION_NAME_CODE_MAP:
        return STATION_NAME_CODE_MAP[station_name]
    return None



# pylint: disable=no-member
"""
Update materival view facilities_all to include
interconnector metadata

Revision ID: fbd4ae450de2
Revises: 0a7eaccf205e
Create Date: 2021-01-13 16:56:13.707337

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "fbd4ae450de2"
down_revision = "0a7eaccf205e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
        drop materialized view if exists mv_facility_all;
        create materialized view mv_facility_all as
        select
            fs.trading_interval,
            f.code,
            f.fueltech_id,
            f.network_id,
            f.network_region,
            f.interconnector,
            f.interconnector_region_to,
            max(fs.energy) as energy,
            case when avg(bs.price) >= 0  and min(fs.energy) >= 0 then
                coalesce(
                    max(fs.energy) * avg(bs.price),
                    0.0
                )
            else 0.0
            end as market_value,
            case when avg(f.emissions_factor_co2) >= 0  and min(fs.energy) >= 0 then
                coalesce(
                    max(fs.energy) * avg(f.emissions_factor_co2),
                    0.0
                )
            else 0.0
            end as emissions
        from mv_facility_energy_hour fs
            left join facility f on fs.facility_code = f.code
            left join balancing_summary bs on
                bs.trading_interval = fs.trading_interval
                and bs.network_id=f.network_id
                and bs.network_region = f.network_region
        where
            f.fueltech_id is not null
        group by
            1,
            f.code,
            f.fueltech_id,
            f.network_id,
            f.network_region,
            f.interconnector,
            f.interconnector_region_to
        order by 1 desc;
    """
    )


def downgrade() -> None:
    pass



# pylint: disable=no-member
"""
Data fix demand

Revision ID: 31ea8ecdfa44
Revises: 156aec569e8d
Create Date: 2021-01-31 00:28:35.164479

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "31ea8ecdfa44"
down_revision = "156aec569e8d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("update balancing_summary set demand_total=NULL where network_id='NEM';")


def downgrade() -> None:
    pass



# @TODO crawl
#
import scrapy

from opennem.pipelines.npi.facilities import NPIStoreFacility

NPI_GEOJSON_URL = "https://data.gov.au/geoserver/npi/wfs?request=GetFeature&typeName=ckan_043f58e0_a188_4458_b61c_04e5b540aea4&outputFormat=json"


class NPIFacilitySpider(scrapy.Spider):
    # name = "au.npi.facilities"

    start_urls = [NPI_GEOJSON_URL]

    pipelines = set([NPIStoreFacility])

    def parse(self, response):
        yield {"records": response.json()}



import logging
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Dict, Optional

import tomlkit
from packaging.version import LegacyVersion, parse

logger = logging.getLogger(__name__)


# get metadata from pyproject.toml
def get_project_meta() -> Optional[Dict]:
    pyproject = Path(__file__).parent / "../../pyproject.toml"

    if not pyproject.is_file():
        return None

    with pyproject.open() as pp_file:
        file_contents = pp_file.read()

    project_meta = None

    try:
        project_meta = tomlkit.parse(file_contents)
    except Exception as e:
        logger.error("Error parsing project metadata: {}".format(e))
        return None

    return project_meta


# get version from VERSION file
def get_pkg_version() -> Optional[str]:
    import pkgutil

    pkg_data = None

    try:
        pkg_data = pkgutil.get_data(__package__, "VERSION")
    except Exception:
        pass

    if not pkg_data:
        return None

    version = pkg_data.decode("utf-8").strip()

    del pkgutil

    return version


pkg_meta = get_project_meta()
version = None

if pkg_meta:
    CUR_YEAR = datetime.now().year
    poetry_meta = pkg_meta["tool"]["poetry"]
    project = str(poetry_meta["name"])
    copyright = f"{CUR_YEAR}, OpenNEM"
    version = str(poetry_meta["version"])

release = version


class VersionPart(Enum):
    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"


def get_version(version_part: Optional[VersionPart] = None) -> str:
    if not version:
        return ""

    version_parsed = parse(version)

    if not version_part:
        return version

    if isinstance(version_parsed, LegacyVersion):
        return version

    if version_part == VersionPart.MAJOR:
        return str(version_parsed.major)

    if version_part == VersionPart.MINOR:
        return str(version_parsed.minor)

    if version_part == VersionPart.PATCH:
        return str(version_parsed.minor)

    return version



# pylint: disable=no-member
"""
Net interchange on balancing summary per region

Revision ID: 1057033fe1ea
Revises: 405b2ea7b351
Create Date: 2021-01-21 16:45:46.555735

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1057033fe1ea"
down_revision = "405b2ea7b351"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "balancing_summary",
        sa.Column("net_interchange", sa.Numeric(), nullable=True),
    )


def downgrade() -> None:
    op.drop_column("balancing_summary", "net_interchange")



# pylint: disable=no-member
"""
Update unique index mv_facility_all

Revision ID: f048f1d8922c
Revises: 53521de3619b
Create Date: 2021-01-28 18:24:47.305094

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "f048f1d8922c"
down_revision = "53521de3619b"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "CREATE UNIQUE INDEX IF NOT EXISTS idx_mv_facility_unique ON mv_facility_all (trading_interval, network_id, code);"
    )


def downgrade() -> None:
    op.execute("drop index idx_mv_facility_unique;")



from opennem.pipelines.wem.pulse import WemStorePulse
from opennem.spiders.wem import WemCurrentSpider


class WemPulseCurrentSpider(WemCurrentSpider):
    name = "au.wem.live.pulse"
    pipelines_extra = set([WemStorePulse])

    start_url = (
        "https://aemo.com.au/aemo/data/wa/infographic/neartime/pulse.csv"
    )



# pylint: disable=no-member
"""
Fix pollution values

Revision ID: 64987ea01b57
Revises: 10f179b31ef8
Create Date: 2020-11-17 18:34:20.197091

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "64987ea01b57"
down_revision = "10f179b31ef8"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "update facility set emissions_factor_co2 = NULL where network_id='WEM'"
    )


def downgrade():
    pass



__all__ = ["dumps", "loads"]


alphabet = "ZAC2B3EF4GH5TK67P8RS9WXY"


def dumps(number: int) -> str:
    """"""
    if not isinstance(number, int):
        raise TypeError("number must be an integer")

    if number < 0:
        return "-" + dumps(-number)

    value = ""

    while number != 0:
        number, index = divmod(number, len(alphabet))
        value = alphabet[index] + value

    return value or "0"


def loads(value: str) -> int:
    """"""
    if len(value) > 13:
        raise ValueError("base24 input too large")

    return int(value, 24)



import logging
import math
from datetime import date, datetime, timedelta, timezone
from typing import Generator, List, Optional, Tuple, Union

from dateutil.parser import ParserError, parse

from opennem.api.stats.schema import ScadaDateRange
from opennem.core.networks import NetworkSchema
from opennem.core.normalizers import normalize_whitespace
from opennem.utils.timezone import UTC, is_aware, make_aware

logger = logging.getLogger(__name__)

# Date formats
# See: https://docs.python.org/3.8/library/datetime.html#strftime-and-strptime-behavior
DATE_FORMATS = [
    "%Y/%m/%d,%H:%M:%S",
    "%Y%m%d",
    #
    "%d/%m/%y %H:%M",
    "%Y/%m/%d %H:%M:%S",
    "%d/%m/%Y %I:%M:%S %p",  # excel format
    "%Y%m%d%H%M%S",  # bom format
]

DATE_CURRENT = datetime.now()
DATE_CURRENT_YEAR = DATE_CURRENT.year


def optimized_data_parser(date_str: str) -> datetime:
    """
    Turns out that dateutil's date parser is slow since
    it does a lot of string parsing. Here we try matching
    the date using known string formats.
    """
    dt_return = None

    date_str = normalize_whitespace(date_str).strip()

    for date_format_str in DATE_FORMATS:
        try:
            dt_return = datetime.strptime(date_str, date_format_str)
        except (KeyError, OverflowError, ValueError, AttributeError):
            continue
        if dt_return:
            return dt_return

    return dt_return


def parse_date(
    date_str: Union[str, datetime],
    date_format: Optional[str] = None,
    network: Optional[NetworkSchema] = None,
    dayfirst: bool = True,
    yearfirst: bool = False,
    is_utc: bool = False,
    timezone: timezone = None,
    use_optimized: bool = True,
) -> Optional[datetime]:
    dt_return = None

    if isinstance(date_str, datetime):
        dt_return = date_str

    elif isinstance(date_str, str):
        # avoid strptime if we can
        try:
            dt_return = datetime.fromisoformat(date_str.replace("/", "-"))
        except ValueError:
            pass

        if not dt_return and date_format:
            dt_return = datetime.strptime(date_str, date_format)

        if not dt_return and use_optimized:
            dt_return = optimized_data_parser(date_str)

        if not dt_return:
            try:
                dt_return = parse(date_str, dayfirst=dayfirst, yearfirst=yearfirst)
            except ParserError:
                raise ValueError("Invalid date string passed")

    else:
        raise ValueError("Require a datetime or string object to parse date")

    if network:
        tz = network.get_timezone()

        if tz:
            if is_aware(dt_return):
                if hasattr(tz, "localize"):
                    dt_return = tz.localize()
                else:
                    dt_return = dt_return.replace(tzinfo=tz)
            else:
                dt_return = make_aware(dt_return, timezone=tz)

    if is_utc:
        tz = UTC

        if dt_return and is_aware(dt_return):
            if tz and hasattr(tz, "localize"):
                dt_return = tz.localize()
            else:
                dt_return = dt_return.replace(tzinfo=tz)
        else:
            dt_return = make_aware(dt_return, timezone=tz)

    if timezone:
        dt_return = make_aware(dt_return, timezone=timezone)

    return dt_return


def date_series(
    start: Union[datetime, date] = None,
    end: Union[datetime, date] = None,
    length: int = 30,
    interval: timedelta = timedelta(days=1),
    reverse: bool = False,
) -> List[datetime]:
    """
    Generate a datetime series

    """
    if start and isinstance(start, datetime):
        start = start.date()

    if end and isinstance(end, datetime):
        end = end.date()

    if not start:
        start = datetime.now().date()

    if end:
        length = int(abs((end - start) / interval))

    next_record = start

    for _ in range(length):
        yield next_record

        if reverse:
            next_record -= interval
        else:
            next_record += interval


total_months = lambda dt: dt.month + 12 * dt.year


def total_weeks(d1: Union[datetime, date], d2: Union[datetime, date]) -> int:
    monday1 = d1 - timedelta(days=d1.weekday())
    monday2 = d2 - timedelta(days=d2.weekday())

    return abs(int((monday2 - monday1).days / 7))


def month_series(
    start: Union[datetime, date],
    end: Union[datetime, date],
    length: Optional[int] = None,
    reverse: bool = False,
) -> Generator[datetime, None, None]:
    """
    Generate a series of months
    """
    step = 1

    if end < start:
        step = -1

    if start == end:
        yield datetime(start.year, start.month, 1)

    for tot_m in range(total_months(start) - 1, total_months(end) - 2, step):
        y, m = divmod(tot_m, 12)
        yield datetime(y, m + 1, 1)


def week_series(
    start: Union[datetime, date],
    end: Union[datetime, date],
    length: Optional[int] = None,
) -> Generator[Tuple[int, int], None, None]:
    """
    Generate week series M -> S
    """
    reverse = False

    if end < start:
        reverse = True

    length = total_weeks(start, end)

    for week_i in range(0, length):
        if reverse:
            cur_date = start - timedelta(weeks=week_i)
        else:
            cur_date = start + timedelta(weeks=week_i)

        cur_cal = cur_date.isocalendar()

        yield cur_cal[0], cur_cal[1]


def date_range_from_week(
    year: int, week: int, network: Optional[NetworkSchema] = None
) -> ScadaDateRange:
    """
    Get a scada date range from week number with
    network awareness
    """
    start_date_str = f"{year}-W{week - 1}-1"
    start_date_dt = datetime.strptime(start_date_str, "%Y-W%W-%w")

    end_date = start_date_dt + timedelta(days=7)

    scada_range = ScadaDateRange(start=start_date_dt, end=end_date, network=network)

    return scada_range


def chop_microseconds(delta: timedelta) -> timedelta:
    return delta - timedelta(microseconds=delta.microseconds)


def get_date_component(format_str: str, dt: datetime = None) -> str:
    """
    Get the format string part out of a date

    ex.
    >>> get_date_component("%Y")
    > 2020
    """
    if dt:
        return dt.strftime(format_str)
    return datetime.now().strftime(format_str)


def subtract_week(subject: datetime) -> datetime:
    return subject - timedelta(days=7)


def subtract_days(subject: Optional[datetime] = None, days: int = 30) -> datetime:
    if not subject:
        subject = datetime.now(timezone.utc)

    return subject - timedelta(days=days)


def is_valid_isodate(date: str, check_timezone: bool = False) -> bool:
    """Check if a string is a valid ISO formatted datestring"""
    dt = None

    try:
        dt = datetime.fromisoformat(date)
    except ValueError:
        return False

    if check_timezone:
        if dt.tzinfo:
            return True
        else:
            return False

    return True


def get_quarter(dt: datetime) -> int:
    """Get the quarter from a datetime"""
    return math.ceil(dt.month / 3.0)


def get_end_of_last_month(dt: datetime) -> datetime:
    """Get the end of the previous month"""
    dtn = dt.replace(day=1) - timedelta(days=1)

    return dtn



from scrapy import Spider


class CERRegistrySpider(Spider):
    # name = "au.cer.registry"
    start_urls = [
        "https://www.rec-registry.gov.au/rec-registry/app/public/power-station-register"
    ]



# pylint: disable=no-member
"""
Set continous aggregate policies

Revision ID: 4a0f8c654257
Revises: 47b8cabac4bb
Create Date: 2021-01-23 18:57:00.102509

"""
from alembic import op
from sqlalchemy.exc import ProgrammingError

# revision identifiers, used by Alembic.
revision = "4a0f8c654257"
down_revision = "47b8cabac4bb"
branch_labels = None
depends_on = None

VIEWS = [
    {"name": "mv_nem_facility_power_5min", "interval": "5 minutes"},
    {"name": "mv_wem_facility_power_30min", "interval": "5 minutes"},
    {"name": "mv_interchange_power_nem_region", "interval": "5 minutes"},
    {"name": "mv_interchange_energy_nem_region", "interval": "5 minutes"},
    {"name": "mv_facility_energy_hour", "interval": "30 minutes"},
    {"name": "mv_balancing_summary_region_hour", "interval": "30 minutes"},
]

query_drop = "SELECT remove_continuous_aggregate_policy('{view_name}')"

query_create = """SELECT add_continuous_aggregate_policy('{view_name}',
    start_offset => INTERVAL '1 day',
    end_offset => NULL,
    schedule_interval => INTERVAL '{schedule_size}');
"""


def upgrade() -> None:
    for v in VIEWS:
        q_drop = query_drop.format(view_name=v["name"])
        q_create = query_create.format(view_name=v["name"], schedule_size=v["interval"])

        with op.get_context().autocommit_block():
            try:
                op.execute(q_drop)
            except ProgrammingError:
                pass

        with op.get_context().autocommit_block():
            op.execute(q_create)


def downgrade() -> None:
    pass



# pylint: disable=no-member
"""
Create index mv_facility_all_trading_interval

Revision ID: e161acce7727
Revises: 98e2968b63e4
Create Date: 2021-01-18 18:29:47.808544

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "e161acce7727"
down_revision = "98e2968b63e4"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "create index if not exists idx_mv_facility_all_trading_interval_day_aest on mv_facility_all (date_trunc('day'::text, timezone('AEST'::text, trading_interval)) DESC, fueltech_id);"
    )


def downgrade() -> None:
    op.execute("drop index idx_mv_facility_all_trading_interval_day_aest;")



"""
OpenNEM Bulk Insert Pipeline

Bulk inserts records using temporary tables and CSV imports with copy_from

This is by far the fastest way to bulk insert large records sets of consistent width
and supports on conflict upserts with postgres

This should be the last step of the pipeline and requires `RecordsToCSVPipeline`
from `csv.py`

@TODO use pydantic throughout to tidy up the schema that is passed through the pipeline

"""
import logging
from datetime import datetime
from io import StringIO
from typing import Any, List, Optional, Union

# from sqlalchemy.exc import StatementError
from sqlalchemy.sql.schema import Column, Table

from opennem.db import get_database_engine
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)

BULK_INSERT_QUERY = """
    CREATE TEMP TABLE __tmp_{table_name}_{tmp_table_name}
    (LIKE {table_name} INCLUDING DEFAULTS)
    ON COMMIT DROP;

    COPY __tmp_{table_name}_{tmp_table_name} FROM STDIN WITH (FORMAT CSV, HEADER TRUE, DELIMITER ',');

    INSERT INTO {table_name}
        SELECT *
        FROM __tmp_{table_name}_{tmp_table_name}
    ON CONFLICT {on_conflict}
"""

BULK_INSERT_CONFLICT_UPDATE = """
    ({pk_columns}) DO UPDATE set {update_values}
"""


def build_insert_query(
    table: Table,
    update_cols: List[Union[str, Column]] = None,
) -> str:
    """
    Builds the bulk insert query
    """
    on_conflict = "DO NOTHING"

    def get_column_name(column: Union[str, Column]) -> str:
        if isinstance(column, Column) and hasattr(column, "name"):
            return column.name
        if isinstance(column, str):
            return column.strip()
        return ""

    update_col_names = []

    if update_cols:
        update_col_names = [get_column_name(c) for c in update_cols]

    update_col_names = list(filter(lambda c: c, update_col_names))

    primary_key_columns = [c.name for c in table.__table__.primary_key.columns.values()]

    if len(update_col_names):
        on_conflict = BULK_INSERT_CONFLICT_UPDATE.format(
            pk_columns=",".join(primary_key_columns),
            update_values=", ".join([f"{n} = EXCLUDED.{n}" for n in update_col_names]),
        )

    query = BULK_INSERT_QUERY.format(
        table_name=table.__table__.name,
        on_conflict=on_conflict,
        tmp_table_name=datetime.strftime(datetime.now(), "%Y%m%d%H%M%S"),
    )

    logger.debug(query)

    return query


class BulkInsertPipeline(object):
    @check_spider_pipeline
    def process_item(self, item: List[dict], spider):
        num_records = 0
        conn = get_database_engine().raw_connection()

        if not isinstance(item, list):
            spider_name = "unknown"

            if spider and hasattr(spider, "name"):
                spider_name = spider.name

            logger.error(
                "Did not get a list of items in bulk inserter from spider {}".format(spider_name)
            )

            return {"num_records": "ERROR"}

        for single_item in item:
            if "csv" not in single_item:
                logger.error("No csv record passed to bulk inserter")
                return 0

            csv_content: StringIO = single_item["csv"]

            if "table_schema" not in single_item:
                logger.error("No table model passed to bulk inserter")
                return item

            table: Table = single_item["table_schema"]

            update_fields: Optional[List[Union[str, Column[Any]]]] = None

            if "update_fields" in single_item:
                update_fields = single_item["update_fields"]

            sql_query = build_insert_query(table, update_fields)

            try:
                cursor = conn.cursor()
                cursor.copy_expert(sql_query, csv_content)
                conn.commit()
            except Exception as generic_error:
                if hasattr(generic_error, "hide_parameters"):
                    generic_error.hide_parameters = True
                logger.error(generic_error)

            try:
                num_records += len(csv_content.getvalue().split("\n")) - 1
            except Exception:
                pass

        return {"num_records": num_records}



"""
OpenNEM Time Series Schema

Defines a tiem series and methods to extract start and end times based on max/min from
SCADA or other ranes and the start/end used in SQL queries

All ranges and queries are _inclusive_ of start/end dates

End is the most recent time chronoligally ordered:

[start] === series === [end]

"""

from __future__ import annotations

from datetime import date, datetime
from typing import Optional, Union

from datetime_truncate import truncate as date_trunc

from opennem.api.time import human_to_interval, human_to_period
from opennem.schema.core import BaseConfig
from opennem.schema.network import NetworkSchema
from opennem.schema.time import TimeInterval, TimePeriod
from opennem.utils.dates import get_end_of_last_month
from opennem.utils.interval import get_human_interval
from opennem.utils.version import CUR_YEAR


def valid_trunc(trunc: str) -> str:
    get_human_interval(trunc)
    return trunc


class DatetimeRange(BaseConfig):
    """Sime start/end date used for sql queries"""

    start: datetime

    # Alias last
    end: datetime

    interval: TimeInterval

    @property
    def trunc(self) -> str:
        return self.interval.interval_human

    @property
    def length(self) -> int:
        """Return the number of buckets in the range"""
        _comp = self.start
        _delta = get_human_interval(self.trunc)
        count = 1

        while _comp < self.end:
            _comp += _delta
            count += 1

        return count


class TimeSeries(BaseConfig):
    """A time series. Consisting of a start and
    end date (inclusive), a bucket size, timezone info
    and methods to extract DatetimeRange's for sql queries
    """

    # Start and end dates
    start: datetime
    end: datetime

    # The network for this date range
    # used for timezone and offsets
    network: NetworkSchema

    # The interval which provides the bucket size
    interval: TimeInterval

    # The length of the series to extract
    period: TimePeriod

    # extract a particular year
    year: Optional[int]

    # Forecast means the time series goes forward from now
    forecast: bool = False

    # Default forward forecast time period
    forecast_period: str = "7d"

    def get_range(self) -> DatetimeRange:
        """Return a DatetimeRange from the time series for queries"""
        start = self.start
        end = self.end

        # If its a forward looking forecast
        # jump out early
        if self.forecast:
            start = self.end
            end = self.end + get_human_interval(self.forecast_period)

            start = start.astimezone(self.network.get_fixed_offset())
            end = end.astimezone(self.network.get_fixed_offset())

            return DatetimeRange(start=start, end=end, interval=self.interval)

        # subtract the period (ie. 7d from the end for start if not all)
        if self.period == human_to_period("all"):
            start = date_trunc(start, self.interval.trunc)

            # If its all per month take the end of the last month
            if self.interval == human_to_interval("1M"):
                end = date_trunc(get_end_of_last_month(end), "day")

            self.year = None

        else:
            start = self.end - get_human_interval(self.period.period_human)

        if self.year:
            if self.year > end.year:
                raise Exception("Specified year is great than end year")

            start = start.replace(year=self.year, month=1, day=1)
            start = date_trunc(start, self.interval.trunc)

            if self.year != CUR_YEAR:
                end = datetime(year=self.year, month=12, day=31, hour=0, minute=0, second=0)
            else:
                end = date_trunc(end, "day")

        # localize times
        start = start.astimezone(self.network.get_fixed_offset())
        end = end.astimezone(self.network.get_fixed_offset())

        dtr = DatetimeRange(start=start, end=end, interval=self.interval)

        return dtr



# pylint: disable=no-member
"""
Material view energy network

Revision ID: ae504bc24d95
Revises: 674d602ecf59
Create Date: 2020-12-14 19:09:27.495897

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "ae504bc24d95"
down_revision = "674d602ecf59"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
    drop view if exists mv_network_energy_hour cascade;

    CREATE OR REPLACE VIEW  mv_network_energy_hour WITH (timescaledb.continuous) AS
    select
        time_bucket('1 hour', fs.trading_interval) as trading_interval,
        fs.network_id,
        energy_sum(fs.generated, '1 hour') * interval_size('1 hour', count(fs.generated)) as energy
    from facility_scada fs
    group by
        1,
        2;

    """
    )


def downgrade() -> None:
    op.execute("drop view if exists mv_network_energy_hour cascade;")



"""
Schemas for stats
"""

from datetime import datetime
from enum import Enum
from typing import Any, List

from pydantic import BaseModel
from pydantic.class_validators import validator
from xlrd.xldate import xldate_as_datetime

from opennem.core.normalizers import clean_float


class StatTypes(Enum):
    CPI = "CPI"
    Inflation = "INFLATION"

    def __str__(self) -> str:
        return "%s" % self.value


class StatDatabase(BaseModel):
    stat_date: datetime
    country: str = "au"
    stat_type: StatTypes
    value: float

    @validator("value", pre=True)
    def parse_cpi_value(cls, value: Any) -> float:
        v = clean_float(value)

        if not v:
            raise ValueError("No Value")

        return v

    class Config:
        orm_mode = True
        anystr_strip_whitespace = True
        use_enum_values = True


class StatsSet(BaseModel):
    name: str
    source_url: str
    fetched_date: datetime
    stats: List[StatDatabase]


class AUCpiData(BaseModel):
    quarter_date: datetime
    cpi_value: float

    @validator("quarter_date", pre=True)
    def parse_quarter_date(cls, value) -> datetime:
        v = xldate_as_datetime(value, 0)

        if not v or not isinstance(v, datetime):
            raise ValueError("Invalid CPI quarter")

        return v

    @validator("cpi_value", pre=True)
    def parse_cpi_value(cls, value: Any) -> float:
        v = clean_float(value)

        if not v:
            raise ValueError("No CPI Value")

        return v


class AUInflationData(BaseModel):
    quarter_date: datetime
    inflation_value: float

    @validator("quarter_date", pre=True)
    def parse_quarter_date(cls, value) -> datetime:
        v = xldate_as_datetime(value, 0)

        if not v or not isinstance(v, datetime):
            raise ValueError("Invalid CPI quarter")

        return v

    @validator("inflation_value", pre=True)
    def parse_cpi_value(cls, value: Any) -> float:
        v = clean_float(value)

        if not v:
            raise ValueError("No inflation Value")

        return v



from functools import reduce
from typing import List, Optional

from opennem.core.normalizers import is_single_number


def getcommonletters(strlist):
    return "".join(
        [
            x[0]
            for x in zip(*strlist)
            if reduce(lambda a, b: (a == b) and a or None, x)
        ]
    )


def findcommonstart(strlist):
    strlist = strlist[:]
    prev = None
    while True:
        common = getcommonletters(strlist)
        if common == prev:
            break
        strlist.append(common)
        prev = common

    return getcommonletters(strlist)


def station_code_from_duids(duids: List[str]) -> Optional[str]:
    """
        Derives a station code from a list of duids

        ex.

        BARRON1,BARRON2 => BARRON
        OSBAG,OSBAG => OSBAG

    """
    if type(duids) is not list:
        return None

    if not duids:
        return None

    if len(duids) == 0:
        return None

    duids_uniq = list(set(duids))

    common = findcommonstart(duids_uniq)

    if not common:
        return None

    # strip last character if we have one
    if is_single_number(common[-1]):
        common = common[:-1]

    if common.endswith("_"):
        common = common[:-1]

    if len(common) > 2:
        return common

    return None



# pylint: disable=no-member
"""
Update policy mv_nem_facility_power_5min

Revision ID: f7e87ce05302
Revises: 99a6fee86d14
Create Date: 2021-01-26 01:59:53.604797

"""
from alembic import op
from sqlalchemy.exc import ProgrammingError

# revision identifiers, used by Alembic.
revision = "f7e87ce05302"
down_revision = "99a6fee86d14"
branch_labels = None
depends_on = None

VIEWS = [
    {"name": "mv_nem_facility_power_5min", "interval": "5 minutes", "start_offset": "2 hours"},
]

query_drop = "SELECT remove_continuous_aggregate_policy('{view_name}')"

query_create = """SELECT add_continuous_aggregate_policy('{view_name}',
    start_offset => INTERVAL '{start_offset}',
    end_offset => NULL,
    schedule_interval => INTERVAL '{schedule_size}');
"""


def upgrade() -> None:
    for v in VIEWS:
        q_drop = query_drop.format(view_name=v["name"])
        q_create = query_create.format(
            view_name=v["name"], schedule_size=v["interval"], start_offset=v["start_offset"]
        )

        with op.get_context().autocommit_block():
            try:
                op.execute(q_drop)
            except ProgrammingError:
                pass

        with op.get_context().autocommit_block():
            op.execute(q_create)


def downgrade() -> None:
    pass



import logging

from opennem.core.loader import load_data

logger = logging.getLogger(__name__)


def load_statuses() -> dict:
    statuses = load_data("facility_status.json", from_fixture=True)

    status_dict = {}

    for s in statuses:
        _code = s.get("code", None)
        status_dict[_code] = s

    return status_dict


FACILITY_STATUS_FIXTURE = load_statuses()


def map_aemo_facility_status(facility_status: str) -> str:
    """
        Maps an AEMO facility status to an Opennem facility status


    """

    unit_status = facility_status.lower().strip()

    if unit_status.startswith("in service"):
        return "operating"

    if unit_status.startswith("in commissioning"):
        return "commissioning"

    if unit_status.startswith("committed"):
        return "committed"

    if unit_status.startswith("maturing"):
        return "maturing"

    if unit_status.startswith("emerging"):
        return "emerging"

    raise Exception(
        "Could not find AEMO status for facility status: {}".format(
            unit_status
        )
    )


def parse_facility_status(status_code: str) -> dict:
    if status_code not in FACILITY_STATUS_FIXTURE.keys():
        raise Exception("Invalid facility status: {}".format(status_code))

    return FACILITY_STATUS_FIXTURE[status_code]



# pylint: disable=no-member
"""
Balancing summary region mv

Revision ID: d0089ded02af
Revises: 2f198fa45c76
Create Date: 2021-01-14 02:00:44.434001

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "d0089ded02af"
down_revision = "2f198fa45c76"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
    drop view if exists mv_balancing_summary_region_hour cascade;

    CREATE OR REPLACE VIEW mv_balancing_summary_region_hour WITH (timescaledb.continuous) AS
    select
        time_bucket('1 hour', bs.trading_interval) as trading_interval,
        bs.network_id,
        bs.network_region,
        avg(bs.price) as price
    from balancing_summary bs
    group by
        1, 2, 3;

    """
    )


def downgrade() -> None:
    op.execute("drop view if exists mv_balancing_summary_region_hour cascade;")



import csv
import logging
from datetime import datetime

from sqlalchemy.exc import IntegrityError

from opennem.core.facility_code import parse_wem_facility_code
from opennem.core.fueltechs import lookup_fueltech
from opennem.core.normalizers import (
    clean_capacity,
    normalize_duid,
    normalize_string,
    participant_name_filter,
    station_name_cleaner,
)
from opennem.db import SessionLocal
from opennem.db.models.opennem import Facility, Location, Participant, Station
from opennem.schema.network import NetworkWEM
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class WemStoreFacility(object):
    @check_spider_pipeline
    def process_item(self, item, spider):
        s = SessionLocal()

        records_added = 0
        csvreader = csv.DictReader(item["content"].split("\n"))

        for row in csvreader:

            if "Participant Code" not in row:
                logger.error("Invalid row")
                continue

            if row["Facility Type"] in [
                "Demand Side Program",
                "Non-Dispatchable Load",
                "Network",
            ]:
                continue

            participant = None

            participant_name = participant_name_filter(row["Participant Name"])
            participant_network_name = normalize_string(
                row["Participant Name"]
            )
            participant_code = normalize_duid(row["Participant Code"])

            participant = (
                s.query(Participant)
                .filter(Participant.code == participant_code)
                .one_or_none()
            )

            if not participant:
                participant = Participant(
                    created_by=spider.name,
                    approved_at=datetime.now(),
                    code=participant_code,
                    name=participant_name,
                    network_name=participant_network_name,
                )
                s.add(participant)
                s.commit()

                logger.debug(
                    "Participant not found created new database entry: %s",
                    participant_code,
                )

            station = None
            facility = None

            facility_code = normalize_duid(row["Facility Code"])
            station_code = parse_wem_facility_code(facility_code)

            station = (
                s.query(Station)
                .filter(Station.code == station_code)
                .one_or_none()
            )

            if not station:
                station = Station(
                    created_by="opennem.wem.facilities",
                    approved_at=datetime.now(),
                    code=station_code,
                    network_code=station_code,
                    participant=participant,
                )

                location = Location(state="WA")

                station.location = location

                logger.debug("Added WEM station: {}".format(station_code))

            facility = (
                s.query(Facility)
                .filter(Facility.code == facility_code)
                .one_or_none()
            )

            if not facility:
                facility = Facility(
                    created_by=spider.name,
                    approved_at=datetime.now(),
                    code=facility_code,
                    network_id="WEM",
                    network_code=facility_code,
                    network_region="WEM",
                )

            capacity_registered = clean_capacity(row["Maximum Capacity (MW)"])
            capacity_unit = clean_capacity(row["Maximum Capacity (MW)"])
            registered_date = row["Registered From"]

            facility.status_id = "operating"
            facility.capacity_registered = capacity_registered
            facility.unit_id = 1
            facility.unit_number = 1
            facility.unit_capacity = capacity_unit

            if registered_date:
                registered_date_dt = datetime.strptime(
                    registered_date, "%Y-%m-%d %H:%M:%S"
                )
                facility.registered = registered_date_dt

            facility.station = station

            s.add(facility)
            records_added += 1

        try:
            s.commit()
        except IntegrityError as e:
            logger.error(e)
            pass
        except Exception as e:
            logger.error("Error: {}".format(e))
        finally:
            s.close()

        return records_added


class WemStoreLiveFacilities(object):
    def parse_interval(self, date_str):
        dt = datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")
        dt_aware = NetworkWEM.get_timezone().localize(dt)

        return dt_aware

    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = SessionLocal()

        records_added = 0
        csvreader = csv.DictReader(item["content"].split("\n"))

        for row in csvreader:

            if "PARTICIPANT_CODE" not in row:
                logger.error("Invalid row")
                continue

            participant = None

            participant_code = normalize_duid(row["PARTICIPANT_CODE"])

            participant = (
                s.query(Participant)
                .filter(Participant.code == participant_code)
                .one_or_none()
            )

            if not participant:
                participant = Participant(
                    created_by=spider.name,
                    approved_by=spider.name,
                    approved_at=datetime.now(),
                    code=participant_code,
                    network_code=participant_code,
                    network="WEM",
                )

                s.add(participant)
                s.commit()

                logger.warning(
                    "Participant not found created new database entry: {}".format(
                        participant_code
                    )
                )

            station = None
            facility = None

            facility_code = normalize_duid(row["FACILITY_CODE"])
            station_code = parse_wem_facility_code(facility_code)
            station_name = station_name_cleaner(row["DISPLAY_NAME"])
            station_network_name = normalize_string(row["DISPLAY_NAME"])

            station = (
                s.query(Station)
                .filter(Station.code == station_code)
                .one_or_none()
            )

            if not station:
                station = Station(
                    created_by=spider.name,
                    approved_by=spider.name,
                    approved_at=datetime.now(),
                    code=station_code,
                    network_code=station_code,
                    participant=participant,
                )

                location = Location(state="WA")
                s.add(location)

                station.location = location

                logger.debug("Added WEM station: {}".format(station_code))

            lat = row["LATITUDE"]
            lng = row["LONGITUDE"]

            station.name = station_name
            station.network_name = station_network_name

            if lat and lng and not station.location.geom:
                station.location.geom = "SRID=4326;POINT({} {})".format(
                    lat, lng
                )
                station.location.geocode_by = "aemo"
                station.location.geocode_approved = True

            facility = (
                s.query(Facility)
                .filter(Facility.code == facility_code)
                .one_or_none()
            )

            if not facility:
                facility = Facility(
                    created_by="opennem.wem.live.facilities",
                    approved_at=datetime.now(),
                    network_id="WEM",
                    code=facility_code,
                    network_code=facility_code,
                    network_region="WEM",
                )

            registered_date = row["YEAR_COMMISSIONED"]

            if registered_date:
                registered_date_dt = None

                date_fmt = "%Y"

                try:
                    registered_date_dt = datetime.strptime(
                        registered_date, date_fmt
                    )
                except Exception:
                    logger.error(
                        "Bad date: %s for format %s", registered_date, date_fmt
                    )

                if registered_date_dt:
                    facility.registered = registered_date_dt

            fueltech = lookup_fueltech(
                fueltype=row["PRIMARY_FUEL"], techtype=row["FACILITY_TYPE"]
            )

            if fueltech and not facility.fueltech:
                facility.fueltech_id = fueltech

            facility.status_id = "operating"
            facility.station = station

            s.add(station)
            s.add(facility)
            s.commit()
            records_added += 1

        try:
            s.commit()
        except IntegrityError as e:
            logger.error(e)
            pass
        except Exception as e:
            logger.error("Error: {}".format(e))
        finally:
            s.close()

        return records_added



import decimal
from typing import List, Optional

from sqlalchemy.orm.session import Session

from opennem.db.models.opennem import Facility, Location, Station


def get_stations(
    session: Session,
    only_approved: bool = True,
    name: Optional[str] = None,
    limit: Optional[int] = None,
    page: int = 1,
) -> List[Station]:
    """
        API controller that gets all stations sorted and joined

    """

    stations = (
        session.query(Station)
        .join(Station.facilities)
        .join(Station.location)
        .join(Facility.fueltech)
        .filter(Facility.fueltech_id.isnot(None))
        .filter(Facility.status_id.isnot(None))
    )

    if name:
        stations = stations.filter(Station.name.like("%{}%".format(name)))

    if only_approved:
        stations = stations.filter(Station.approved.is_(True))

    stations = stations.order_by(
        Facility.network_region,
        Station.name,
        Facility.network_code,
        Facility.code,
    )

    if limit:
        stations = stations.limit(limit)

    stations = stations.all()

    return stations



import logging
import os
from io import BytesIO
from zipfile import ZipFile

from requests import RequestException

from opennem.utils.handlers import _handle_zip, chain_streams, open
from opennem.utils.http import http
from opennem.utils.mime import decode_bytes, mime_from_content, mime_from_url
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


def _fallback_download_handler(url: str) -> bytes:
    """
    This was previously a fallback download handler
    but the name is redundant as it's now the primary
    and takes precedence over the legacy downloader


    """
    r = http.get(url)

    if not r.ok:
        raise Exception("Bad link returned {}: {}".format(r.status_code, url))

    content = BytesIO(r.content)

    file_mime = mime_from_content(content)

    if not file_mime:
        file_mime = mime_from_url(url)

    # @TODO handle all this in utils/archive.py
    # and make it all generic to handle other
    # mime types
    if file_mime == "application/zip":
        with ZipFile(content) as zf:
            if len(zf.namelist()) == 1:
                return zf.open(zf.namelist()[0]).read()

            c = []
            stream_count = 0

            for filename in zf.namelist():
                if filename.endswith(".zip"):
                    c.append(_handle_zip(zf.open(filename), "r"))
                    stream_count += 1
                else:
                    c.append(zf.open(filename))

            return chain_streams(c).read()

    return content.getvalue()


class LinkExtract(object):
    """
    Parse and extracts links in items.

    If the pipeline item has a "link" it'll download the content
    and attach it to the item as "content" along with some
    basic metadata.

    """

    @check_spider_pipeline
    def process_item(self, item, spider):
        if "link" not in item:
            return item

        url = item["link"]
        fh = None
        content = None
        _, file_extension = os.path.splitext(url)

        try:
            _bytes_obj = _fallback_download_handler(url)
            content = decode_bytes(_bytes_obj)
        except Exception as e:
            logger.error(e)

        if content:
            item["content"] = content
            item["extension"] = file_extension
            return item

        try:
            logger.info("Grabbing: {}".format(url))
            fh = open(url)
        except RequestException:
            logger.info("Bad link: {}".format(url))
        except Exception as e:
            logger.error("Error: {}".format(e))

        if fh:
            content = fh.read()

            item["content"] = content
            item["extension"] = file_extension
            return item
